﻿(*
  Ita IDE Plugin

  Copyright (c) 2014-2015 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit UItaMain;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Generics.Collections,
  System.IniFiles, System.Win.Registry, System.Rtti, System.TypInfo,
  Vcl.Controls, Vcl.Graphics, Vcl.Menus, Vcl.Imaging.jpeg, Vcl.Imaging.pngimage,
  Vcl.GraphUtil, ToolsAPI, UItaConfig;

procedure LoadBackgroundImage;
procedure SaveSettings;

type
  TImageAlign = (iaLeftTop, iaLeftBottom, iaRightTop, iaRightBottom);

  TImageSettings = record
    FileName: string;
    Align: TImageAlign;
    Position: TPoint;
    Scale: Integer;
    AlphaBlendValue: Byte;
  end;

const
  SCALE_BASE = 1000;

var
  ImageSettings: TImageSettings = (
    FileName: '';
    Align: iaRightTop;
    Position: (X: 0; Y: 0);
    Scale: SCALE_BASE;
    AlphaBlendValue: 64
  );

implementation

const
  sEVFillRect = '@Editorcontrol@TCustomEditControl@EVFillRect$qqrrx18System@Types@TRect';
  sEVScrollRect = '@Editorcontrol@TCustomEditControl@EVScrollRect$qqrp18System@Types@TRectt1ii';
  sEditControlList = '@Editorcontrol@EditControlList';

  EVFillRectCodes: array[0..5] of Byte = (
    $53,      // PUSH EBX
    $56,      // PUSH ESI
    $8B, $F2, // MOV ESI,EDX
    $8B, $D8  // MOV EBX,EAX
  );
  EVScrollRectCodes: array[0..5] of Byte = (
    $55,          // PUSH EBP
    $8B, $EC,     // MOV EBP,ESP
    $83, $C4, $FF // ADD ESP,-xx
  );

  CoreIdeModuleName = {$IFDEF VER230}'coreide160.bpl'{$ENDIF} // XE2
                      {$IFDEF VER240}'coreide170.bpl'{$ENDIF} // XE3
                      {$IFDEF VER250}'coreide180.bpl'{$ENDIF} // XE4
                      {$IFDEF VER260}'coreide190.bpl'{$ENDIF} // XE5
                      {$IFDEF VER270}'coreide200.bpl'{$ENDIF} // XE6
                      {$IFDEF VER280}'coreide210.bpl'{$ENDIF} // XE7
                      {$IFDEF VER290}'coreide220.bpl'{$ENDIF} // XE8
                      ;

  HighlightRegKey =
    {$IFDEF VER230}'\Software\Embarcadero\BDS\9.0\Editor\Highlight\'{$ENDIF}  // XE2
    {$IFDEF VER240}'\Software\Embarcadero\BDS\10.0\Editor\Highlight\'{$ENDIF} // XE3
    {$IFDEF VER250}'\Software\Embarcadero\BDS\11.0\Editor\Highlight\'{$ENDIF} // XE4
    {$IFDEF VER260}'\Software\Embarcadero\BDS\12.0\Editor\Highlight\'{$ENDIF} // XE5
    {$IFDEF VER270}'\Software\Embarcadero\BDS\14.0\Editor\Highlight\'{$ENDIF} // XE6
    {$IFDEF VER280}'\Software\Embarcadero\BDS\15.0\Editor\Highlight\'{$ENDIF} // XE7
    {$IFDEF VER290}'\Software\Embarcadero\BDS\16.0\Editor\Highlight\'{$ENDIF} // XE8
    ;
  HighlightRegName = 'Background Color New';

var
  TCustomEditControl_EVScrollRect: Pointer;
  TCustomEditControl_EVFillRect: Pointer;
  EditControlList: ^TList;

  ImageSize: TSize;
  LeftGutterProp: ppropinfo;
  BufferDict: TObjectDictionary<TColor,TBitmap>;
  TempBuffer: TBitmap;

type
  TEditControl = class(TCustomControl);

procedure EditControlOnResize(Self: TObject; Sender: TEditControl);
begin
  Sender.Invalidate;
end;

const
  EditControlOnResizeEvent: TMethod = (Code: @EditControlOnResize; Data: nil);

procedure HookEVFillRect(Self: TEditControl; const R: TRect);
var
  leftGutterWidth: Integer;
  imageRect, editorRect: TRect;
  buf: TBitmap;
begin
  if not Assigned(Self.OnResize) then
    Self.OnResize := TNotifyEvent(EditControlOnResizeEvent);

  leftGutterWidth := 0;
  case ImageSettings.Align of
    iaLeftTop:
      editorRect := Bounds(ImageSettings.Position.X, ImageSettings.Position.Y, ImageSize.cx, ImageSize.cy);
    iaLeftBottom:
      editorRect := Bounds(ImageSettings.Position.X, Self.ClientHeight - ImageSettings.Position.Y - ImageSize.cy, ImageSize.cx, ImageSize.cy);
    iaRightTop:
      editorRect := Bounds(Self.ClientWidth - ImageSettings.Position.X - ImageSize.cx, ImageSettings.Position.Y, ImageSize.cx, ImageSize.cy);
    iaRightBottom:
      editorRect := Bounds(Self.ClientWidth - ImageSettings.Position.X - ImageSize.cx, Self.ClientHeight - ImageSettings.Position.Y - ImageSize.cy, ImageSize.cx, ImageSize.cy);
  end;
  if (ImageSettings.Align in [iaLeftTop, iaLeftBottom]) and (LeftGutterProp <> nil) then
  begin
    leftGutterWidth := GetOrdProp(Self, LeftGutterProp);
    editorRect.Offset(leftGutterWidth, 0);
  end;

  if not IntersectRect(editorRect, editorRect, R) then
  begin
    Self.Canvas.FillRect(R);
    Exit;
  end;

  imageRect := editorRect;
  case ImageSettings.Align of
    iaLeftTop:
      imageRect.Offset(-ImageSettings.Position.X-leftGutterWidth, -ImageSettings.Position.Y);
    iaLeftBottom:
      imageRect.Offset(-ImageSettings.Position.X-leftGutterWidth, -(Self.ClientHeight - ImageSettings.Position.Y - ImageSize.cy));
    iaRightTop:
      imageRect.Offset(-(Self.ClientWidth - ImageSettings.Position.X - ImageSize.cx), -ImageSettings.Position.Y);
    iaRightBottom:
      imageRect.Offset(-(Self.ClientWidth - ImageSettings.Position.X - ImageSize.cx), -(Self.ClientHeight - ImageSettings.Position.Y - ImageSize.cy));
  end;

  if editorRect = R then
  begin
    if BufferDict.TryGetValue(Self.Canvas.Brush.Color, buf) then
      Self.Canvas.CopyRect(editorRect, buf.Canvas, imageRect)
    else
      FillRect(Self.Canvas.Handle, R, Self.Canvas.Brush.Handle);
  end
  else begin
    if (TempBuffer.Width < R.Width) or (TempBuffer.Height < R.Height) then
      TempBuffer.SetSize(R.Width, R.Height);

    TempBuffer.Canvas.Brush.Color := Self.Canvas.Brush.Color;
    TempBuffer.Canvas.FillRect(Rect(0, 0, R.Width, R.Height));

    if BufferDict.TryGetValue(Self.Canvas.Brush.Color, buf) then
    begin
      TempBuffer.Canvas.CopyRect(Bounds(editorRect.Left-R.Left, editorRect.Top-R.Top, editorRect.Width, editorRect.Height),
        buf.Canvas, imageRect);
    end;

    Self.Canvas.CopyRect(R, TempBuffer.Canvas, Bounds(0, 0, R.Width, R.Height));
  end;
end;

procedure HookEVScrollRect(Self: TWinControl; const Src, Dest: TRect; DX, DY: Integer);
begin
  Self.Invalidate;
end;

procedure RefreshEditors;
var
  iEditor: Integer;
begin
  if EditControlList = nil then Exit;

  for iEditor := 0 to EditControlList^.Count-1 do
    TWinControl(EditControlList^[iEditor]).Invalidate;
end;

function Hook: Boolean;

  function IsValidCodes(P: PByte; const Codes: array of Byte): Boolean;
  var
    i: Integer;
  begin
    if P = nil then Exit(False);
    for i := Low(Codes) to High(Codes) do
    begin
      if P^ and Codes[i] <> P^ then
        Exit(False);
      Inc(p);
    end;
    Result := True;
  end;

  procedure JmpHookProc(Target: Pointer; ReplaceProc: Pointer);
  var
    oldProtect: Cardinal;
  begin
    VirtualProtect(Target, 5, PAGE_READWRITE, oldProtect);
    PByte(Target)^ := $E9;
    PInteger(Integer(Target)+1)^ := Integer(ReplaceProc) - Integer(Target) - 5;
    VirtualProtect(Target, 5, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

var
  hModule: THandle;
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
begin
  Result := False;

  hModule := GetModuleHandle(CoreIdeModuleName);
  if hModule = 0 then Exit;

  TCustomEditControl_EVFillRect := GetProcAddress(hModule, sEVFillRect);
  TCustomEditControl_EVScrollRect := GetProcAddress(hModule, sEVScrollRect);
  EditControlList := GetProcAddress(hModule, sEditControlList);

  if not IsValidCodes(TCustomEditControl_EVFillRect, EVFillRectCodes) or
     not IsValidCodes(TCustomEditControl_EVScrollRect, EVScrollRectCodes) then
  begin
    TCustomEditControl_EVFillRect := nil;
    TCustomEditControl_EVScrollRect := nil;
    EditControlList := nil;
    Exit;
  end;

  JmpHookProc(TCustomEditControl_EVFillRect, @HookEVFillRect);
  JmpHookProc(TCustomEditControl_EVScrollRect, @HookEVScrollRect);

  typ := ctx.FindType('EditorControl.TEditControl');
  if typ <> nil then
  begin
    prop := typ.GetProperty('LeftGutter');
    if prop <> nil then
      LeftGutterProp := TRttiInstanceProperty(prop).PropInfo;
  end;

  RefreshEditors;

  Result := True;
end;
                                                                                  
procedure Unhook;

  procedure UnhookProc(Target: Pointer; Size: Integer; const OrgCodes: array of Byte);
  var
    oldProtect: Cardinal;
  begin
    VirtualProtect(Target, Size, PAGE_READWRITE, oldProtect);
    Move(OrgCodes[0], Target^, Size);
    VirtualProtect(Target, Size, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

var
  i: Integer;
begin
  if TCustomEditControl_EVFillRect = nil then Exit;
  if TCustomEditControl_EVScrollRect = nil then Exit;
  if EditControlList = nil then Exit;

  if EditControlList^ <> nil then
    for i := 0 to EditControlList^.Count-1 do
      TEditControl(EditControlList^[i]).OnResize := nil;

  UnhookProc(TCustomEditControl_EVFillRect, 5, EVFillRectCodes);
  UnhookProc(TCustomEditControl_EVScrollRect, 5, EVScrollRectCodes);

  RefreshEditors;
end;

procedure LoadSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ExtractFilePath(GetModuleName(HInstance)) + 'itaide.ini');
  try
    ImageSettings.FileName := ini.ReadString('Image', 'FileName', ImageSettings.FileName);
    ImageSettings.Align := TImageAlign(ini.ReadInteger('Image', 'Align', Ord(ImageSettings.Align)));
    ImageSettings.Position.X := ini.ReadInteger('Image', 'Left', ImageSettings.Position.X);
    ImageSettings.Position.Y := ini.ReadInteger('Image', 'Top', ImageSettings.Position.Y);
    ImageSettings.Scale := ini.ReadInteger('Image', 'Scale', ImageSettings.Scale);
    ImageSettings.AlphaBlendValue := ini.ReadInteger('Image', 'Alpha', ImageSettings.AlphaBlendValue);
  finally
    ini.Free;
  end;
end;

procedure SaveSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ExtractFilePath(GetModuleName(HInstance)) + 'itaide.ini');
  try
    ini.WriteString('Image', 'FileName', ImageSettings.FileName);
    ini.WriteInteger('Image', 'Align', Ord(ImageSettings.Align));
    ini.WriteInteger('Image', 'Left', ImageSettings.Position.X);
    ini.WriteInteger('Image', 'Top', ImageSettings.Position.Y);
    ini.WriteInteger('Image', 'Scale', ImageSettings.Scale);
    ini.WriteInteger('Image', 'Alpha', ImageSettings.AlphaBlendValue);
    ini.UpdateFile;
  finally
    ini.Free;
  end;
end;

procedure LoadBackgroundImage;

  procedure LoadImage(Image: TPngImage);
  var
    ext: string;
    png: TPngImage;
    jpg: TJPEGImage;
    bmp: TBitmap;
    i, x, y: Integer;
    TransparencyChunk: TChunktRNS;
    pAlpha: PByteArray;
    alpha: Double;
  begin
    ext := LowerCase(ExtractFileExt(ImageSettings.FileName));
    if (not FileExists(ImageSettings.FileName)) or (ext = '.png') then
    begin
      png := TPNGImage.Create;
      try
        if not FileExists(ImageSettings.FileName) then
          png.LoadFromResourceName(HInstance, 'PNG_DEFAULT')
        else
          png.LoadFromFile(ImageSettings.FileName);
        if png.Header.ColorType = COLOR_RGBALPHA then
        begin
          Image.Assign(png);
          if ImageSettings.AlphaBlendValue <> 255 then
          begin
            alpha := ImageSettings.AlphaBlendValue / 255;
            for y := 0 to Image.Height-1 do
            begin
              pAlpha := Image.AlphaScanline[y];
              for x := 0 to Image.Width-1 do
                pAlpha^[x] := Trunc(pAlpha^[x] * alpha);
            end;
          end;
        end
        else begin
          Image.Assign(png);
          case Image.Header.ColorType of
            COLOR_GRAYSCALE, COLOR_RGB:
            begin
              Image.CreateAlpha;
              for i := 0 to Image.Height-1 do
                FillChar(Image.AlphaScanline[i]^, Image.Width, ImageSettings.AlphaBlendValue);
            end;
            COLOR_PALETTE:
            begin
              Image.CreateAlpha;
              TransparencyChunk := TChunktRNS(Image.Chunks.ItemFromClass(TChunktRNS));
              FillChar(TransparencyChunk.PaletteValues[0], 256, ImageSettings.AlphaBlendValue);
            end;
          end;
        end;
      finally
        png.Free;
      end;
    end
    else if (ext = '.jpg') or (ext = '.jpeg') then
    begin
      jpg := TJPEGImage.Create;
      try
        jpg.LoadFromFile(ImageSettings.FileName);
        png := TPngImage.CreateBlank(COLOR_RGBALPHA, 8, jpg.Width, jpg.Height);
        try
          png.Canvas.Draw(0, 0, jpg);
          for i := 0 to png.Height-1 do
            FillChar(png.AlphaScanline[i]^, png.Width, ImageSettings.AlphaBlendValue);
          Image.Assign(png);
        finally
          png.Free;
        end;
      finally
        jpg.Free;
      end;
    end
    else if ext = '.bmp' then
    begin
      bmp := TBitmap.Create;
      try
        bmp.LoadFromFile(ImageSettings.FileName);

        png := TPngImage.CreateBlank(COLOR_RGBALPHA, 8, bmp.Width, bmp.Height);
        try
          png.Canvas.Draw(0, 0, bmp);
          for i := 0 to png.Height-1 do
            FillChar(png.AlphaScanline[i]^, png.Width, ImageSettings.AlphaBlendValue);
          Image.Assign(png);
        finally
          png.Free;
        end;
      finally
        bmp.Free;
      end;
    end
    else
      raise Exception.Create('Unsupported image format');
  end;

  function GetBackgroundColors: TArray<TColor>;
  var
    reg: TRegistry;
    keys: TStringList;
    i: Integer;
    color: TColor;
  begin
    Result := nil;

    reg := TRegistry.Create;
    try
      reg.RootKey := HKEY_CURRENT_USER;
      if not reg.OpenKeyReadOnly(HighlightRegKey) then Exit;

      keys := TStringList.Create;
      try
        reg.GetKeyNames(keys);
        for i := 0 to keys.Count-1 do
        begin
          if not reg.OpenKeyReadOnly(HighlightRegKey + keys[i]) then Continue;

          color := StringToColor(reg.ReadString(HighlightRegName));
          SetLength(Result, Length(Result)+1);
          Result[High(Result)] := color;
        end;
      finally
        keys.Free;
      end;
    finally
      reg.Free;
    end;
  end;

  function GetBlendedImage(Image: TPngImage; Color: TColor): TBitmap;
  begin
    Result := TBitmap.Create;
    try
      Result.SetSize(Image.Width, Image.Height);
      Result.Canvas.Brush.Color := Color;
      Result.Canvas.FillRect(Rect(0, 0, Result.Width, Result.Height));
      Result.Canvas.Draw(0, 0, Image);
    except
      FreeAndNil(Result);
    end;
  end;

var
  Image: TPngImage;
  buf, buf1: TBitmap;
  color: TColor;
begin
  BufferDict.Clear;
  try
    Image := TPngImage.Create;
    try
      LoadImage(Image);

      ImageSize.cx := Image.Width;
      ImageSize.cy := Image.Height;

      for color in GetBackgroundColors do
      begin
        if BufferDict.ContainsKey(color) then Continue;

        if ImageSettings.Scale = SCALE_BASE then
        begin
          buf := GetBlendedImage(Image, color);
          if buf <> nil then
            BufferDict.Add(color, buf);
        end
        else begin
          buf := GetBlendedImage(Image, color);
          if buf <> nil then
          try
            buf1 := TBitmap.Create;
            try
              ScaleImage(buf, buf1, ImageSettings.Scale / SCALE_BASE);
              buf1.PixelFormat := pfDevice;
              BufferDict.Add(color, buf1);
              ImageSize.cx := buf1.Width;
              ImageSize.cy := buf1.Height;
            except
              buf1.Free;
            end;
          finally
            buf.Free;
          end;
        end;
      end;
    finally
      Image.Free;
    end;
  except
    BufferDict.Clear;
    Exit;
  end;

  RefreshEditors;
end;

var
  ItaIDEMenu: TMenuItem;

procedure ItaIDEImageMenuClick(Self, Sender: TObject);
begin
  ShowConfigDialog;
end;

procedure RegisterMenu;
var
  nta: INTAServices;
  i, j: Integer;
  method: TMethod;
begin
  nta := BorlandIDEServices as INTAServices;
  for i := 0 to nta.MainMenu.Items.Count-1 do
  begin
    if nta.MainMenu.Items[i].Name <> 'ToolsMenu' then Continue;

    for j := 0 to nta.MainMenu.Items[i].Count-1 do
    begin
      if nta.MainMenu.Items[i][j].Name <> 'ToolsToolsItem' then Continue;

      ItaIDEMenu := TMenuItem.Create(nil);
      ItaIDEMenu.AutoHotkeys := maManual;
      if SameText(PreferredUILanguages, 'JA') then
        ItaIDEMenu.Caption := '痛IDE設定'
      else
        ItaIDEMenu.Caption := 'Ita IDE Options';
      method.Code := @ItaIDEImageMenuClick;
      method.Data := ItaIDEMenu;
      ItaIDEMenu.OnClick := TNotifyEvent(method);
      nta.MainMenu.Items[i].Insert(j, ItaIDEMenu);

      Break;
    end;

    Break;
  end;
end;

procedure UnregisterMenu;
begin
  ItaIDEMenu.Free;
end;

initialization
  BufferDict := TObjectDictionary<TColor,TBitmap>.Create([doOwnsValues]);
  LoadSettings;
  LoadBackgroundImage;
  if not Hook then Exit;
  RegisterMenu;
  TempBuffer := TBitmap.Create;
finalization
  UnregisterMenu;
  Unhook;
  BufferDict.Free;
  TempBuffer.Free;
end.
